http = require 'http'
url = require 'url'
querystring = require 'querystring'

exports.setupESI = ->
	handleESI = (req, res, next) ->
		endWait = false		

		# MODIFY res.write
		write = res.write  # store the original function		
		res.write = (chunk, encoding) -> # wrap write to hook into the exit path through the layers
			res.write = write # put the original back
			content_type = 	if res._headers? and res._headers['content-type']? then	res._headers['content-type'] else 'text/plain' # if no content-type is provided, set to 'text/plain'
			unless content_type.indexOf('text') == -1 # only process ESIs if response data contains text
				console.log "\nres.write\nrequest.url: " + req.url + " content-type: " + content_type
				endWait = true
				esiProcessor = new exports.processESI chunk, (err, chunk) ->
					#console.log "processESI callback"
					res.writeHead (res.statusCode ? 200),
					  'Content-Length': chunk.length
					  'Content-Type': content_type
					endWait = false
					res.write chunk, encoding # call the original on callback
					#res.end()
				esiProcessor.request = req.headers
				esiProcessor.request.url = url.parse('http://' + req.headers['host'] + req.url, true)
				esiProcessor.execute()
			else			
				res.write chunk, encoding # call the original				

		# MODIFY res.end
		end = res.end
		res.end = (data, encoding) ->
			res.end = end
			#console.log "\nres.end\n request.url: " + req.url + " wait: " + endWait
			if data
				res.write data, encoding
			unless endWait == true # only end if not waiting (for async)
				res.end()

		next() # pass through to the next layer

exports.processESI = (data, callback) ->
	@data = data.toString()
	@callback = callback
	@request = []
	@request['url'] = url.parse('http://localhost:5678')

	@execute = ->
		console.log "Searching for ESIs..."
		esi_tags = @searchESI @data
		console.log "Found " + esi_tags.length + " ESIs"
		if (esi_tags.length != 0) # there are ESIs to process
			for k of esi_tags
				src_vars = @processESIvars esi_tags[k]
				console.log "Fetching ESI '" + esi_tags[k] + "'"
				@fetchESI [esi_tags[k], src_vars], (src, status, body) =>
					console.log "Fetch ESI '" + src[1] + "' returned status " + status
					if status != 200 then body = '' # remove error messages if http status is not 200
					console.log "ESI tag subsitution for '" + src[0] + "'"				
					remaining = @subESI src[0], body
					console.log remaining + " remaining ESIs"
					if remaining == 0 # no remaining ESIs to process, callback
						console.log "Display Data\n"
						@callback false, @data						
		else
			console.log "Display Data\n"
			@callback false, @data	

	@searchESI = (data) ->
		re = new RegExp('<esi[^>]* src=("[^>"]+"|\'[^>\']+\')[^>]*>(</esi[^>]*>)?', 'gi')
		esi_tags = data.match(re)
		for k of esi_tags
			re = new RegExp(' src=("(.*?)"|\'(.*?)\')')
			match = esi_tags[k].match(re)
			esi_tags[k] = match[2] ? match[3] #quotes surrounding src: 2 - double, 3 - single
		return esi_tags ? []

	@fetchESI = (src, callback) ->
		if (src !instanceof Array) # if src is string, create array
			src = [src,src]
		uri = url.parse(src[1])
		options =
			host: uri.hostname ? @request.url.hostname
			port: uri.port ? (if uri.protocol? then 80 else @request.url.port)
			path: (uri.pathname ? '/') + (uri.search ? '') + (uri.hash ? '')
		http.get(options, (res) ->
			body = ''
			res.on 'data', (chunk) =>
				body += chunk.toString()
			res.on 'end', =>
				callback src, res.statusCode, body
		).on 'error', (e) ->
			callback src, 500, e.message

	@subESI = (src, body) ->
		re = new RegExp('<esi[^>]* src=("' + RegExp.escape(src) + '"|\'' + RegExp.escape(src) + '\')[^>]*>(</esi[^>]*>)?', 'gi')
		@data = @data.replace(re, body.replace(/\$/g,'$$$')) # if story contains $ escape it
		return @searchESI(@data).length # number of ESIs remaining

	@processESIvars = (data) ->
		re = new RegExp('\\$\\([^\\)]+\\)', 'gi')
		esi_vars = data.match(re)
		for i of esi_vars	
			re = new RegExp('\\(([^{\\|\\)]*)({([^}]*)})?(\\|([^\\|{)]*))?\\)', 'i')
			esi_var = esi_vars[i].match(re)
			esi_var_result = @lookupESIvar(esi_var[1], esi_var[3])
			if !esi_var_result and esi_var[5]? # set default if exists
				esi_var_result = esi_var[5]
			data = data.replace(esi_vars[i], esi_var_result)
		data
		
	@lookupESIvar = (esi_var, options) ->
		result = ''
		switch esi_var.toUpperCase()
			when 'HTTP_ACCEPT_LANGUAGE'
				result = @request['accept-language']
			when 'HTTP_COOKIE'
				result = ''
				if options?
					re = new RegExp('(^| )' + options + '=([^;]*)', 'i')
					cookie = @request['cookie'] ? ''
					matches = cookie.match(re)
					if matches?
						result = matches[2]
			when 'HTTP_HEADER'
				result = @request[options]
			when 'HTTP_HOST'
				result = @request['host']
			when 'HTTP_REFERER'
				result = @request['referer']
			when 'HTTP_USER_AGENT'
				result = @request['user-agent']
			when 'QUERY_STRING'
				if options?
					result = @request.url.query[options]
				else
					result = querystring.stringify @request.url.query
		result ? ''

	RegExp.escape = (text) ->
	  text.replace /[-[\]{}()*+?.,\\^$|#\s]/g, "\\$&"

	return